@@ -1,9 +1,13 @@
 
+import collections
+import contextlib
 import csv
 import io
+import itertools
 import os
 import typing
-import urllib, urllib.parse
+import urllib
+import urllib.parse
 import zipfile
 
 try:
@@ -11,6 +15,7 @@ try:
 except ImportError:  # pragma: no cover
     requests = None
 
+import comma.exceptions
 import comma.extras
 import comma.typing
 
@@ -26,13 +31,22 @@ __all__ = [
 
     "DefaultDialect",
 
+    "is_anystr",
     "is_local",
     "is_url",
     "detect_line_terminator",
     "open_stream",
     "open_csv",
 
+    "validate_header",
+    "has_header",
+    "detect_sequence_kind",
+    "detect_sequence_common_kind",
+
     "multislice_sequence",
+    "multislice_range",
+    "multislice_index",
+
     "zip_html_tag",
 ]
 
@@ -53,7 +67,27 @@ class DefaultDialect(csv.Dialect):
     The default dialect for output, when no dialect is provided.
     """
 
+    @classmethod
+    def override(cls, **kwargs) -> csv.Dialect:
+        """
+        Creates a `csv.Dialect` object that only overrides certain
+        settings from the default dialect.
+        """
+        obj = cls()
+
+        for field, value in kwargs.items():
+            if field not in obj.__dict__:
+                raise AttributeError(
+                    "Class `{}` does not have a field `{}` to override".format(
+                        cls, field))
+            obj.__dict__[field] = value
+
+        return obj
+
     def __init__(self):
+        """
+        Creates a `csv.Dialect` with the package's default settings.
+        """
         self.delimiter = ","
         self.doublequote = True
         self.escapechar = "\\"
@@ -381,7 +415,34 @@ def open_stream(
     if type(sample) is bytes:
         if encoding is None:
             encoding = comma.extras.detect_encoding(sample)
-        source = io.TextIOWrapper(source, encoding=encoding)
+
+        encoding_candidates = [encoding]
+        if "utf-8" not in encoding_candidates:
+            encoding_candidates.append("utf-8")
+
+        found_encoding = False
+        source_data = source.read()
+        for encoding in encoding_candidates:
+
+            source_with_encoding = io.TextIOWrapper(io.BytesIO(source_data), encoding=encoding)
+
+            # let's sniff test this encoding!
+            # sometimes the encoding detected just does not work
+            try:
+                source_with_encoding.seek(0)
+                source_with_encoding.read()
+                source_with_encoding.seek(0)
+                source = source_with_encoding
+                found_encoding = True
+                break
+            except UnicodeError:
+                continue
+
+        if not found_encoding:
+            raise comma.exceptions.CommaEncodingException(
+                "no suitable encoding could be found, tried: {}".format(
+                    encoding_candidates)
+            )
     
     # try to add useful metadata
     if internal_name is not None:
@@ -394,13 +455,13 @@ def open_stream(
 
 
 def open_csv(
-        source: comma.typing.SourceType,
-        encoding: str = None,
-        delimiters: typing.Optional[typing.Iterable[str]] = None,
-        no_request: bool = False,
-) -> comma.typing.CommaInfo:
+    source: comma.typing.SourceType,
+    encoding: str = None,
+    delimiters: typing.Optional[typing.Iterable[str]] = None,
+    no_request: bool = False,
+) -> comma.typing.CommaInfoType:
     """
-    Returns a `CommaInfo` typed dictionary containing the data and
+    Returns a `CommaInfoType` typed dictionary containing the data and
     metadata related to a CSV file. The `source` can be actual data,
     a local file path, or a URL; it is possible to provide a stream
     that is compressed using ZIP.
@@ -476,9 +537,144 @@ def open_csv(
     return data
 
 
+def validate_header(value: typing.Any) -> typing.List[str]:
+    """
+    Checks that a value is an iterable of string-like values. And converts
+    to a list of strings. This is to verify user-specified header values.
+    """
+    # verify the value is a list (or can be converted to one if,
+    # say, an iterator)
+
+    import comma.exceptions
+
+    if value is None:
+        raise comma.exceptions.CommaInvalidHeaderException(
+            "the header that is being validated is `None`; this is not a "
+            "valid header, though perhaps a signal to unset a variable"
+        )
+
+    try:
+        header_as_list = list(value)
+    except TypeError as e:
+        raise comma.exceptions.CommaInvalidHeaderException(
+            "type error arose as validating specified header") from e
+
+    # check individual fields
+
+    for subval in header_as_list:
+        if subval is None:
+            raise comma.exceptions.CommaInvalidHeaderException(
+                ("attempting to set a header with invalid `None` field name"
+                 "\nfull header: {header}").format(header=header_as_list))
+
+        if not comma.helpers.is_anystr(subval):
+            raise comma.exceptions.CommaInvalidHeaderException(
+                ("attempting to set a header with invalid field name `{field}`"
+                 "\nfull header: {header}").format(
+                    field=subval,
+                    header=header_as_list))
+
+    header_as_list = list(map(str, header_as_list))
+
+    return header_as_list
+
+
+def has_header(value: typing.Any) -> bool:
+    """
+    Checks whether a `value` has a `header` attribute (that is not `None`).
+    """
+    # noinspection PyBroadException
+    try:
+        return value.header is not None
+    except:
+        return False
+
+
+def is_dict_or_list(value: typing.Any) -> typing.Optional[type]:
+    """
+    Checks whether `value` may be some subtype of a `dict` or of a
+    `list`, or whether `value` can be casted to a `dict` or a `list`.
+    (As specified, when both are possibilities, the `dict` is
+    prioritized.)
+    """
+    if value is None:
+        return
+
+    # 1. may have an attribute `header'?
+    if comma.helpers.has_header(value):
+        with contextlib.suppress(Exception):
+            dict(value)
+            return dict
+
+    # 2. may be a dict subtype?
+    if isinstance(value, dict) or isinstance(value, collections.UserDict):
+        with contextlib.suppress(Exception):
+            dict(value)
+            return dict
+
+    # 3. may be a list subtype?
+    if isinstance(value, list) or isinstance(value, collections.UserList):
+        with contextlib.suppress(Exception):
+            list(value)
+            return list
+
+    # 4. may be an iterator?
+    if isinstance(value, typing.Iterator):
+        # if so, since an iterator can usually only be run once,
+        # make a backup of iterator with `itertools.tee()`
+        value, aux = itertools.tee(value)
+        aux1, aux2 = itertools.tee(aux)
+        with contextlib.suppress(Exception):
+            dict(aux1)
+            return dict
+        with contextlib.suppress(Exception):
+            list(aux2)
+            return list
+
+        # if this did not work for iterator, nothing will, give up!
+        return
+
+    # 5. just try the duck-typing outright
+    with contextlib.suppress(Exception):
+        dict(value)
+        return dict
+    with contextlib.suppress(Exception):
+        list(value)
+        return list
+
+    # 6. no clue
+    return
+
+
+def is_dict_or_list_many(a, b, *others):
+    """
+    Checks whether the arguments are `dict` or `list`, then returns: `dict` if
+    they are all a subtype of `dict`; `list` if some are of subtype `dict` and
+    others `list`; and `None` in other situations.
+    """
+    args = [a, b] + list(others)
+    typ = dict
+    for typp in map(comma.helpers.is_dict_or_list, args):
+        if typp is list:
+            typ = list
+        if typp is None:
+            typ = None
+    return typ
+
+
+def dict_or_list_many(*args):
+    """
+    Returns the input parameters casted to the appropriate common subtype,
+    of either `dict` or `list`.
+    """
+    common_subtype = comma.helpers.is_dict_or_list_many(*args)
+    if common_subtype is not None:
+        return list(map(common_subtype, args))
+
+
 def multislice_sequence(
-        sequence: typing.Sequence[typing.Any],
-        slice_list: typing.List[slice] = None,
+    sequence: typing.Sequence[typing.Any],
+    slice_list: typing.List[slice] = None,
 ) -> typing.Sequence[typing.Any]:
     """
     Returns the sub-sequence obtained from sequentially slicing the
@@ -493,14 +689,37 @@ def multislice_sequence(
     return new_sequence
 
 
+def multislice_range(size: int, slice_list: typing.List[slice] = None) -> range:
+    """
+    Returns the range of indexes that are preserved by a succession of
+    slicing operations on the range [0, size). This makes it easier to
+    recover the original index.
+    """
+    # noinspection PyArgumentList
+    return typing.cast(
+        typ=range,
+        val=comma.helpers.multislice_sequence(range(size), slice_list=slice_list))
+
+
+def multislice_index(index: int, size: int, slice_list: typing.List[slice] = None) -> int:
+    """
+    Returns the original index in the original sequence from the index
+    in the sequence after applying multiple slicing operations. This
+    makes it easier to recover the original index.
+    """
+    return multislice_range(size=size, slice_list=slice_list)[index]
+
+
 def zip_html_tag(
-        data: typing.Iterable,
-        in_pattern: str = "<td style='text-align: left;'>{}</td>",
-        out_pattern: str = "<tr>{}</tr>",
-        indent: int = 0
+    data: typing.Iterable,
+    in_pattern: str = "<td style='text-align: left;'>{}</td>",
+    out_pattern: str = "<tr>{}</tr>",
+    indent: int = 0
 ) -> str:
     """
-
+    Returns the HTML code of a template applied to a Python list; to
+    be used to build the rows of tables, or bullet lists in
+    `_repr_html_()` outputs.
     """
     if type(indent) is not int:
         indent = 0
